from enum import Enum, auto

from typing import Union, Callable
from io import TextIOBase
from pathlib import Path


AddressArg = Union[str, Path, TextIOBase]
StrResolver = Callable[[str], bool]


class AddressType(Enum):
    """Type of source or destination.

    Provides built-in methods to load/write from/to any of the supported
    source and destination types.

    Attributes:
        STRING (int): Load from string or return as string.
        STRING_PATH (int): Load from path represented as a string or write to
            such path.
        PATH (int): Load from path represented as pathlib.Path object or
            write to such path.
        STREAM (int): Load from stream represented as file-like object such as
            sys.stdout, open(), StringIO(); or write to such object.
    """

    STRING = auto()
    STRING_PATH = auto()
    PATH = auto()
    STREAM = auto()


class Address:
    """"""

    @staticmethod
    def read(source: AddressArg, str_resolver: StrResolver = None) -> str:
        # Determine type of the source argument
        source_type = Address._resolve(source, str_resolver)

        # Read from this type of soure using proper mapped method
        data = Address._read_method(source_type)(source)

        return data

    @staticmethod
    def write(
        data: str,
        destination: AddressArg = None,
        str_resolver: StrResolver = None,
    ) -> None:
        # Determine type of the destination argument
        destination_type = Address._resolve(destination, str_resolver)

        # Write to this destination using a proper mapped method
        result = Address._write_method(destination_type)(data, destination)

        return result

    @staticmethod
    def _resolve(
        address: AddressArg, str_resolver: StrResolver = None
    ) -> int:
        """Determine which type the ``data`` argument belongs to.

        Args:
            address (AddressArg): Source or destination to read or write data.
            str_resolver (StrResolver): Checker function that
                determines whether the ``data`` argument represents a string
                containing the configuration or a string containing a path to
                configuration file. Only applied when ``data`` is str.

        Returns:
            AddressType: Type of address contained in ``data`` argument.

        Raises:
            TypeError: When address of the ``data`` argument could not be determined.

        """

        str_resolver = str_resolver or Address._dummy_str_resolver

        if isinstance(address, str):
            if str_resolver(address):
                return AddressType.STRING_PATH
            else:
                return AddressType.STRING
        elif isinstance(address, Path):
            return AddressType.PATH
        elif isinstance(address, TextIOBase):
            return AddressType.STREAM
        elif address is None:
            return AddressType.STRING
        else:
            raise TypeError(
                f"Type {type(address)} of {address} is not supported as address of source or destination"
            )

    @staticmethod
    def _from_string(string: str) -> str:
        return string

    @staticmethod
    def _from_string_path(string: str) -> str:
        result = None
        with open(string, "rt") as f:
            result = f.read()
        return result

    @staticmethod
    def _from_path(path: Path) -> str:
        result = path.read_text()
        return result

    @staticmethod
    def _from_stream(stream: TextIOBase) -> str:
        result = stream.read()
        return result

    @staticmethod
    def _to_string(data: str, string: str) -> str:
        return data

    @staticmethod
    def _to_string_path(data: str, string: str) -> str:
        with open(string, "w") as f:
            f.write(data)
        return data

    @staticmethod
    def _to_path(data: str, path: Path) -> str:
        with path.open("w") as f:
            f.write(data)
        return data

    @staticmethod
    def _to_stream(data: str, stream: TextIOBase) -> str:
        stream.write(data)
        return data

    @staticmethod
    def _dummy_str_resolver(string: str) -> bool:
        return False

    @staticmethod
    def _read_method(typ: AddressType):
        mapping = {
            AddressType.STRING: Address._from_string,
            AddressType.STRING_PATH: Address._from_string_path,
            AddressType.PATH: Address._from_path,
            AddressType.STREAM: Address._from_stream,
        }
        return mapping[typ]

    @staticmethod
    def _write_method(typ: AddressType):
        mapping = {
            AddressType.STRING: Address._to_string,
            AddressType.STRING_PATH: Address._to_string_path,
            AddressType.PATH: Address._to_path,
            AddressType.STREAM: Address._to_stream,
        }
        return mapping[typ]



from functools import wraps

from ilexconf.config import Config
from mapz.methods import traverse

from .address import Address, AddressArg, StrResolver

from typing import Callable, Mapping, Any

LoadCallable = Callable[[AddressArg], str]
DumpCallable = Callable[[Mapping[Any, Any]], str]
PreProcessingCallable = Callable[[Mapping[Any, Any]], Mapping[Any, Any]]


def _dummy_load(string: str) -> Mapping:
    return {"value": string}


def _dummy_pre_processing(data: Mapping) -> Mapping:
    return data


def _dummy_dump(data: Mapping[Any, Any], **kwargs) -> str:
    return str(data)


def reader(
    load: LoadCallable = _dummy_load,
    str_resolver: StrResolver = None,
    pre_processing: PreProcessingCallable = _dummy_pre_processing,
):
    """Decorator for read adapter functions such as ``from_``.

    Args:
        load: Function that loads certain format from string and returns a Mapping.
        str_resolver: Function that determines whether str data passed to decorated function is a path to file or not.
        pre_processing: Function that somehow transforms Mapping obtained from ``load`` before passing it to ``Config`` constructor.
    """

    def decorator_reader(func):
        @wraps(func)
        def wrapper_reader(data: AddressArg, **kwargs):
            # Read data using proper loader
            serialized = Address.read(data, str_resolver)

            # Parse string data to dictionary using provided ``load``
            # callable argument.
            deserialized = load(serialized, **kwargs)

            # Perform any pre processing necessary with the parsed
            # dictionary
            deserialized = pre_processing(deserialized)

            return traverse(deserialized, mapping_type=Config)

        return wrapper_reader

    return decorator_reader


def writer(dump: DumpCallable = _dummy_dump, **default_kwargs):
    """Decorator for write adapter functions such as ``to_``.

    Args:
        dump: Function that serializes mapping to a string.
        **default_kwargs: Default keyword arguments that will be passed to the ``dump`` function.
    """

    def decorator_writer(func):
        @wraps(func)
        def wrapper_writer(
            data: Mapping[Any, Any], destination: AddressArg = None, **kwargs
        ):
            if isinstance(data, Config):
                data = data.to_dict()

            # Merge kwargs passed as argument to adapter on top of
            # default kwargs
            merged_kwargs = default_kwargs.copy()
            merged_kwargs.update(kwargs)

            # Convert mapping to string using provided ``dump`` callable
            # argument.
            serialized = dump(data, **merged_kwargs)

            # Save dumped string to destination.
            Address.write(serialized, destination, str_resolver=None)

            # Always return converted (dumped) string.
            return serialized

        return wrapper_writer

    return decorator_writer



import sys
import os
from io import StringIO
from mapz import Mapz


try:
    # Special case for pytest coverage
    if "_pytest_yaml_test_pyyaml" in os.environ:
        raise ImportError()

    from ruamel.yaml import YAML

    yaml = YAML()
    yaml_load = yaml.load
except ImportError:
    try:
        # Special case for pytest coverage
        if "_pytest_yaml_test_noyaml" in os.environ:
            raise ImportError()

        import yaml

        yaml_load = yaml.safe_load
    except ImportError:
        print("AAAAA")
        yaml = None

from .common.decorators import reader, writer


def from_yaml(*args, **kwargs):
    raise NotImplementedError(
        "ruamel.yaml or PyYaml must be installed in order to use YAML adapter"
    )


def to_yaml(*args, **kwargs):
    raise NotImplementedError(
        "ruamel.yaml or PyYaml must be installed in order to use YAML adapter"
    )


if yaml:

    def _load(data: str):
        # If data is a string in a form of
        # "name: boris" or "name:" then it will
        # be parsed by yaml module to a str instance.
        d = yaml_load(data)

        # Do not accept plain string yaml files
        # as configs. Treat such strings as paths.
        if isinstance(d, str):
            with open(data, "rt") as f:
                d = yaml_load(f)

        return d

    def _dump(data, **kwargs) -> str:
        stream = StringIO()

        yaml.dump(data, stream, **kwargs)
        dumped = stream.getvalue()

        return dumped

    @reader(load=_load)
    def from_yaml():
        """Read data from YAML string, file object or path"""
        pass  # pragma: no cover

    @writer(dump=_dump)
    def to_yaml():
        """Write data to YAML file or convert to YAML string"""
        pass  # pragma: no cover



from ilexconf.config import Config

from .env import from_env, to_env
from .ini import from_ini, to_ini
from .json import from_json, to_json

try:
    from .yaml import from_yaml, to_yaml

    yaml_here = True
except ImportError:  # pragma: no cover
    yaml_here = False
    from_yaml = None
    to_yaml = None


formats = Config()
formats.json = {
    "enabled": True,
    "reader": from_json,
    "writer": to_json,
    "extensions": ["json"],
}

formats.yaml = {"enabled": yaml_here, "extensions": ["yaml", "yml"]}
if yaml_here:  # pragma: no cover
    formats.yaml.reader = from_yaml
    formats.yaml.writer = to_yaml

enabled_formats = [f for f in formats if formats[f].enabled]



import sys
import pytest
from pathlib import Path

from ilexconf.adapters.common.address import AddressType, Address


def test_resolver(settings_json_file_path):
    # File
    with open(settings_json_file_path, "rt") as f:
        assert Address._resolve(f) is AddressType.STREAM

    # Stdout
    assert Address._resolve(sys.stdout) is AddressType.STREAM

    # Path
    assert Address._resolve(Path("settings.json")) is AddressType.PATH

    # String
    assert (
        Address._resolve("settings.json", lambda string: False)
        is AddressType.STRING
    )

    # String Path
    assert (
        Address._resolve("settings.json", lambda string: True)
        is AddressType.STRING_PATH
    )

    # None
    assert Address._resolve(None) is AddressType.STRING

    # Unknown argument
    with pytest.raises(TypeError):
        Address._resolve(132)



from ilexconf.config import Config
from ilexconf.adapters.common.decorators import (
    _dummy_load,
    _dummy_dump,
    _dummy_pre_processing,
    reader,
    writer,
)


@reader(
    load=lambda string: {"name": string},
    str_resolver=lambda string: False,
    pre_processing=lambda mapping: {
        k: v for k, v in list(mapping.items()) + [("surname", "Britva")]
    },
)
def from_dummy():
    pass


@writer(dump=lambda mapping: str(mapping))
def to_dummy():
    pass


def test_dummy_load():
    s = "name"
    assert _dummy_load(s) == {"value": s}


def test_dummy_dump():
    d = {"name": "Boris"}
    assert _dummy_dump(d) == str(d)


def test_dummy_pre_processing():
    d = {"name": "Boris"}
    assert _dummy_pre_processing(d) == d


def test_reader_example():
    assert from_dummy("Boris") == Config(
        {"name": "Boris", "surname": "Britva"}
    )


def test_writer_example():
    assert to_dummy({"name": "Boris"}) == str({"name": "Boris"})



from importlib import reload
import os
import sys
import pytest

import ilexconf.adapters.yaml
from ilexconf.adapters.yaml import from_yaml, to_yaml
from ilexconf.adapters.json import from_json


@pytest.fixture(scope="module", autouse=True)
def yamldir(fixture_dir):
    return os.path.join(fixture_dir, "yaml")


def test_read_example(yamldir):

    from ilexconf.adapters.yaml import from_yaml, to_yaml

    config = from_yaml(os.path.join(yamldir, "simple.yaml"))

    json_counterpart_path = os.path.join(yamldir, "simple.json")
    json_analogue = from_json(json_counterpart_path)
    assert config == json_analogue


def test_write_example(yamldir, tmp_path):

    from ilexconf.adapters.yaml import from_yaml, to_yaml

    config = from_yaml(os.path.join(yamldir, "simple.yaml"))

    path = tmp_path / "simple.yaml"
    # import ilexconf.tests.debug
    to_yaml(config, path)

    reloaded = from_yaml(path)
    assert config == reloaded


def test_pyyaml(yamldir, monkeypatch, tmp_path):

    monkeypatch.setenv("_pytest_yaml_test_pyyaml", "true")
    import importlib

    importlib.reload(ilexconf.adapters.yaml)
    config = ilexconf.adapters.yaml.from_yaml(
        os.path.join(yamldir, "simple.yaml")
    )

    json_counterpart_path = os.path.join(yamldir, "simple.json")
    json_analogue = from_json(json_counterpart_path)
    assert config == json_analogue

    path = tmp_path / "simple-pyyaml.yaml"
    ilexconf.adapters.yaml.to_yaml(config, path)

    reloaded = ilexconf.adapters.yaml.from_yaml(path)
    assert config == reloaded


def test_noyaml(yamldir, monkeypatch):
    monkeypatch.setenv("_pytest_yaml_test_pyyaml", "true")
    monkeypatch.setenv("_pytest_yaml_test_noyaml", "true")
    import importlib

    importlib.reload(ilexconf.adapters.yaml)

    with pytest.raises(NotImplementedError):
        ilexconf.adapters.yaml.from_yaml(os.path.join(yamldir, "simple.yaml"))

    with pytest.raises(NotImplementedError):
        ilexconf.adapters.yaml.to_yaml("somepath")



from json import loads, dumps
from typing import Mapping, Any

from .common.decorators import reader, writer


def _load(string) -> Mapping[str, Any]:
    parsed = loads(string)
    return parsed


def _dump(data: Mapping[Any, Any], **kwargs) -> str:
    dumped = dumps(data, **kwargs)
    return dumped


@reader(
    load=_load,
    str_resolver=lambda string: "{" not in string or "}" not in string,
)
def from_json():
    """Read JSON from string, file or path.

    JSON adapter relies on Python's
    `json <https://docs.python.org/3/library/json.html>`_ module when
    working with ``.json`` files.

    Args:
        data: Source of JSON. Accepts file, path or string:

    Returns:
        Config: Returns constructed Config object with the values from JSON source.

    Raises:
        UnsupportedDataSourceType: If ``data`` argument is of wrong type.

    Examples:
        Read from file object:

        >>> from ilexconf import from_json
        >>> with open("path/to/file", "r") as f:
        >>>     config = from_json(f)

        Read from Path object:

        >>> path = pathlib.Path("path/to/file")
        >>> config = from_json(path)

        Read from JSON string:

        >>> config = from_json('{ "name": "Boris" }')

        Read from plain string path:

        >>> config = from_json("path/to/file")

    """
    pass  # pragma: no cover


@writer(dump=_dump, indent=2)
def to_json():
    """Write data to JSON file or convert to JSON string

    JSON adapter relies on Python's
    `json <https://docs.python.org/3/library/json.html>`_ module when
    working with ``.json`` files.

    Args:
        data (Mapping): Mapping object to dump to JSON.
        destination (optional): Accepts either File (stream), Path or str. Defaults to ``None``.
        indent (:obj:`int`, optional): Indentation for JSON. Defaults to 2.
            Use ``indent=None`` to convert without indentation.
        **kwargs: Any other keyword arguments accepted by
            `json.dumps() <https://docs.python.org/3/library/json.html#json.dumps>`_ method.

    Returns:
        str: Returns ``data`` Mapping converted to JSON string.

    Raises:
        UnsupportedDataDestinationType: When ``destination`` argument is of wrong type.

    Examples:
        Write to file object:

        >>> from ilexconf import to_json
        >>> with open("path/to/file", "w") as f:
        >>>     to_json(config, f)

        Write to stream:

        >>> to_json(config, sys.stdout)

        Write to Path object:

        >>> to_json(config, pathlib.Path("path/to/file"))

        Write to string path:

        >>> to_json(config, "path/to/file")

        Convert to string:

        >>> string = to_json(config)

    """
    pass  # pragma: no cover



from ilexconf.config import Config
from ilexconf.adapters import (
    from_json,
    to_json,
    from_ini,
    to_ini,
    from_env,
    to_env,
    from_yaml,
    to_yaml,
)

__myall__ = [
    Config,
    from_json,
    to_json,
    from_ini,
    to_ini,
    from_env,
    to_env,
    from_yaml,
    to_yaml,
]



import sys
from pathlib import Path

from ilexconf.adapters.common.address import Address


def test_read_stream(settings_json_file_path, settings_json_string):
    with open(settings_json_file_path, "rt") as f:
        assert Address.read(f) == settings_json_string


def test_read_string(settings_json_string):
    assert Address.read(settings_json_string) == settings_json_string


def test_read_string_path(settings_json_file_path, settings_json_string):
    assert (
        Address.read(
            settings_json_file_path, str_resolver=lambda string: True
        )
        == settings_json_string
    )


def test_read_path(settings_json_file_path, settings_json_string):
    path = Path(settings_json_file_path)
    assert Address.read(path) == settings_json_string


def test_write_stream(settings_json_string, capsys):
    Address.write(settings_json_string, sys.stdout)
    captured = capsys.readouterr()
    assert captured.out == settings_json_string


def test_write_file(settings_json_string, tmp_path):
    path = tmp_path / "save_file.txt"
    with open(path, "w") as f:
        Address.write(settings_json_string, f)

    with open(path, "r") as f:
        assert f.read() == settings_json_string


def test_write_string(settings_json_string):
    assert Address.write(settings_json_string) == settings_json_string


def test_write_string_path(settings_json_string, tmp_path):
    path = tmp_path / "save_string_path.txt"
    Address.write(
        settings_json_string, str(path), str_resolver=lambda string: True
    )

    with open(str(path), "rt") as f:
        assert f.read() == settings_json_string


def test_write_path(settings_json_file_path, settings_json_string, tmp_path):
    path = Path(tmp_path / "save_path.txt")
    Address.write(settings_json_string, path)

    with path.open("rt") as f:
        assert f.read() == settings_json_string



from mapz import Mapz

from typing import (
    Any,
    Dict,
    Hashable,
    Mapping,
)


class Config(Mapz):
    def __init__(self, *mappings: Mapping[Hashable, Any], **kwargs: Dict):
        super().__init__(*mappings)

        for k, v in kwargs.items():
            self.set(
                k,
                v,
                key_prefix="",
                key_sep="_",
            )

    def __repr__(self) -> str:
        return f"Config{dict.__repr__(self)}"



import pytest


def test_features():
    from ilexconf import Config
    config = Config()

    config.a.b.c = True

    assert (
        config.a.b.c == True
    )
    assert (
        config["a"]["b"]["c"] == True
    )
    assert (
        config["a.b.c"] == True
    )
    assert (
        config["a.b"].c == True
    )
    assert (
        config.a["b.c"] == True
    )
    assert (
        config.a["b"].c == True
    )

    if True:
        config.a.b.c = [
            "my_string",
            {"d": "nested_value"}
        ]

    assert (
        config.a.b.c[0] == "my_string"
    )
    assert (
        config.a.b.c[1] == Config({"d": "nested_value"})
    )
    assert (
        dict(config.a.b.c[1]) == {"d": "nested_value"}
    )



import os

from ilexconf.config import Config

from typing import Mapping, Any, Union


def from_env(prefix="", separator="__"):
    """Read config from environment variables."""

    prefix = "" if prefix is None else prefix
    config = Config()

    for k, v in os.environ.items():
        # Convert current key-value pair to Mapping
        config.set(k, v, key_prefix=prefix, key_sep=separator)

    return config


def to_env(data: Config, prefix="", separator="__"):
    """Export mapping to environment variables.

    This will set the environment variable both for current python
    context as well as for child processes invoed via os.system(),
    popen(), fork(), and execv().

    :param data: Mapping object to export.
    :type data: Mapping[Any, Any]
    """

    config = Config(data)

    # Convert entire hierarchy to flat keys with values.
    # { "database": { "connection": { "host": "1.2.3.4" }}} will turn into
    # { "my_database__connection__host": "1.2.3.4" } with prefix == "my_"
    # and separator == "__".
    flat = config.flatten()

    for k, v in flat.items():
        key = prefix + k.replace(".", separator)

        # Set environment variable for current context
        os.environ[key] = v
        os.putenv(key, v)



from configparser import ConfigParser
from typing import Mapping, Any
from io import StringIO

from ilexconf.config import Config
from mapz import Mapz

from .common.address import AddressArg
from .common.decorators import reader, writer


def _load(string: str) -> Mapping[str, Any]:
    parser = ConfigParser()
    parser.read_string(string)

    d = {}
    for s in parser.sections() + ["DEFAULT"]:
        d[s] = {}
        for k in parser[s]:
            d[s][k] = parser[s][k]
    return d


def _str_resolver(string: str) -> bool:
    return "[" not in string or not "]" in string


def _pre_process(data: Mapping[str, Any]) -> Mapping[str, Any]:
    return data


def _dump(data: Mapping[str, Any]) -> str:
    # No matter what mapping was passed as input we convert it to Config
    config = Config(data)

    # Initialize INI config
    ini = ConfigParser()
    for s in config:
        if isinstance(config[s], Mapz):
            ini[s] = config[s].flatten()
        else:
            ini["DEFAULT"][s] = str(config[s])

    # Serialize INI config to string
    output = StringIO()
    ini.write(output)
    return output.getvalue()


@reader(load=_load, str_resolver=_str_resolver, pre_processing=_pre_process)
def from_ini(source: AddressArg):
    """Read INI from string, file or path.

    INI adapter relies on Python's
    `ConfigParser <https://docs.python.org/3/library/configparser.html>`_ when
    working with ``.ini`` configuration files.

    Args:
        source (AddressArg): Source of INI. Accepts file, path or string:

    Returns:
        Config: Returns constructed Config object with the values from INI source.

    Raises:
        TypeError: If ``source`` argument is of wrong type.

    Examples:
        Read from file object:

        >>> from ilexconf import from_ini
        >>> with open("path/to/file", "r") as f:
        >>>     config = from_ini(f)

        Read from Path object:

        >>> path = pathlib.Path("path/to/file")
        >>> config = from_ini(path)

        Read from INI containing string:

        >>> config = from_ini("name = Boris")

        Read from plain string path:

        >>> config = from_ini("path/to/file")

    """
    pass  # pragma: no cover


@writer(dump=_dump)
def to_ini():
    """Write data to INI file or convert to INI string

    INI adapter relies on Python's
    `ConfigParser <https://docs.python.org/3/library/configparser.html>`_ when
    working with ``.ini`` configuration files.

    Any top level key of the ``data`` argument is considered to be a section.
    Any subsequent values or nested mappings of each of the top level keys are
    flattened before being written.

    Args:
        data (Mapping): Mapping object to dump to INI.
        destination (optional): Accepts either File (stream), Path or str. Defaults to ``None``.

    Returns:
        str: Returns ``data`` Mapping converted to INI string.

    Raises:
        UnsupportedDataDestinationType: When ``destination`` argument is of wrong type.

    Examples:
        Write to file object:

        >>> from ilexconf import to_ini
        >>> with open("path/to/file", "w") as f:
        >>>     to_ini(config, f)

        Write to stream:

        >>> to_ini(config, sys.stdout)

        Write to Path object:

        >>> to_ini(config, pathlib.Path("path/to/file"))

        Write to string path:

        >>> to_ini(config, "path/to/file")

        Convert to string:

        >>> string = to_ini(config)

    """
    pass  # pragma: no cover



import os
import pytest

from ilexconf.config import Config
from ilexconf.adapters.ini import from_ini, to_ini
from ilexconf.adapters.json import from_json


@pytest.fixture(scope="module", autouse=True)
def ini_dir(fixture_dir):
    return os.path.join(fixture_dir, "ini")


@pytest.fixture(scope="module", autouse=True)
def sample_ini(ini_dir):
    return os.path.join(ini_dir, "sample.ini")


def test_read_example(ini_dir, sample_ini):
    config = from_ini(sample_ini)

    json_counterpart_path = os.path.join(ini_dir, "sample.json")
    json_analogue = from_json(json_counterpart_path)
    json_analogue["bitbucket.org"].submerge(json_analogue["DEFAULT"])
    json_analogue["topsecret.server.com"].submerge(json_analogue["DEFAULT"])
    assert config == json_analogue


def test_write_example(sample_ini, tmp_path):
    config = from_ini(sample_ini)

    path = tmp_path / "sample.ini"
    to_ini(config, path)

    reloaded = from_ini(path)
    assert config == reloaded


def test_write_nonstandard(tmp_path):
    config = Config(
        {"database": {"host": "localhost", "port": 1234}, "name": "Duhast"}
    )

    path = tmp_path / "nonstandard.ini"
    to_ini(config, path)

    reloaded = from_ini(path)
    other = {
        "database": {"host": "localhost", "port": "1234", "name": "Duhast"},
        "DEFAULT": {"name": "Duhast"},
    }
    assert reloaded.to_dict() == other
